/* Routines imlementing ELVIS */
/* See elvisexample.g to see how to use this file */
/* The main routine the user should call is calc_lnL_elvis at the end of this file */

/* general naming conventions:
un: unobservables
ob: observables
theta: parameter vector
nuis: nuissance parameters (profiled over)
gam: tilting parameter gamma (see paper for details)
*/

/* hooks to enable parallel operation (default values for serial operation included below) */
proc(1)=oxmpi_bcast(val,root);
  retp(val);
endp;

proc(1)=oxmpi_ismaster();
  retp(1);
endp;

proc(1)=oxmpi_comm_size();
  retp(1);
endp;

proc(1)=oxmpi_comm_rank();
  retp(0);
endp;
/* end of hooks */

/* calculate average of moment function over the unobservables for each individual */
/* myf: pointer to moment function
   ob: the whole sample of observables (each of the observables are in a row, individuals are stacked in cols)
   theta: parameter vector, including nuisance parameters
   gam: tilting parameter gamma
   rep: array controling replications:
    rep[1]: number of equilibration steps
    rep[2]: number of averaging steps
   myseed: random number seed
*/
proc(1)=avg_mom(myf,ob,theta,gam,rep,myseed);
  local myf:proc;
  local dimf,i,avg,num,den,r,obi;
  local un,try_un,dens,try_dens,weight;
  local a;
  local nproc,iproc,j;

  nproc=oxmpi_comm_size();
  iproc=oxmpi_comm_rank();

  dimf=rows(myf(guess_un(ob[1,.]',theta,gam),ob[1,.]',theta));
  avg=zeros(rows(ob),dimf);
  i=1;
  do while (i<=rows(ob));
   if (i%nproc==iproc);
    rndseed(myseed+i);
    num=zeros(dimf,1);
    den=0;
    obi=ob[i,.]';
    un=guess_un(obi,theta,gam);
    r=-rep[1]+1;
    a=zeros(dimf,1);
    do while (r<=rep[2]);
      try_un=jump_un(un,obi,theta);
      try_dens=exp(gam'*myf(try_un,obi,theta)-gam'*myf(un,obi,theta))*rho(try_un,obi,theta)/rho(un,obi,theta);
      if ( rndu(1,1) < (try_dens) );
        un=try_un;
      endif;
      if (r>0);
        a=a+myf(un,obi,theta);
      endif;
      r=r+1;
    endo;
    avg[i,.]=a'/rep[2];
   endif;
   i=i+1;
  endo;
  i=1;
  do while (i<=rows(ob));
    j=1;
    do while (j<=dimf);
      avg[i,j]=oxmpi_bcast(avg[i,j],i%nproc);
      j=j+1;
    endo;
    i=i+1;
  endo; 
  avg=missrv(avg,0);
  retp(avg);
endp;

/* EL */

/* define Owen's de-singularized ln function */
proc(1)=ln_star(x,n);
  if (x>1/n);
    retp(ln(x));
  else;
    retp( -ln(n)-1.5+2*n*x-((n*x)^2)/2 );
  endif;
endp;

/* and 1st derivative */
proc(1)=dln_star(x,n);
  if (x>1/n);
    retp(1/x);
  else;
    retp( 2*n - x*(n^2) );
  endif;
endp;

/* and 2nd derivative */
proc(1)=d2ln_star(x,n);
  if (x>1/n);
    retp(-1/(x^2));
  else;
    retp( -n^2 );
  endif;
endp;

/* calculate the empirical likelihood or the efficient GMM objective function (if DOGMM is defined) */
/* g: values of the moment function in rows, individuals are stacked in cols
   tol: convergence criterion */
proc(1)=calc_lnL_el(g,tol);
#ifdef DOGMM
  local V,gm,n;
  n=rows(g);
  gm=meanc(g);
  V=g'g/n-gm*gm';
  if (ismiss(invswp(V))); retp(-100); endif;
  retp(-gm'*invswp(V)*gm);
#else
  local n,l,h,d,i,k,gi,c,invh,like,old_lnlike,sum,it,bad,lnlike;
  n=rows(g);
  k=cols(g);
  l=zeros(k,1);
  old_lnlike=-1000;
  it=0;
  do while (1);
    h=zeros(k,k);
    d=zeros(k,1);
    lnlike=0;
    bad=0;
    i=1;
    do while (i<=n);
      gi=g[i,.]';
      h=h-gi*gi'*d2ln_star(1-l'gi,n);
      d=d+gi*dln_star(1-l'gi,n);
      lnlike=lnlike-ln_star(1-l'gi,n);
      if (l'gi>1);
        bad=1;
      endif;
      i=i+1;
    endo;
    h=h/n;
    d=d/n;
    lnlike=lnlike/n;
    invh=invswp(h);
    l=l-invh*d;
    if (ismiss(l));
       break;
    endif;
    if (abs(old_lnlike-lnlike)<tol); break; endif;
    old_lnlike=lnlike;
    it=it+1;
  endo;
  if (ismiss(lnlike)); lnlike=old_lnlike; endif;
  retp(lnlike);
#endif
endp;

/* Nelder-Mead simplex "amoeba" algorithm */

/* internal utility function */
proc(4)=amotry(p,y,psum,ihi,fac,funk,data1,data2);
  local funk:proc;
  local fac1,fac2,ytry,ptry,ndim;
  ndim=rows(p);
  ptry=zeros(ndim,1);
  fac1=(1.0-fac)/ndim;
  fac2=fac1-fac;
  ptry=psum*fac1-p[.,ihi]*fac2;
  ytry=funk(ptry,data1,data2);
  if (ytry < y[ihi]);
    y[ihi]=ytry;
    psum = psum + ptry-p[.,ihi];
    p[.,ihi]=ptry;
  endif;
  retp (ytry,p,y,psum);
endp;

/* other internal utility function */
proc(1)=amoeba_loop(p,y,ftol,funk,data1,data2);
  local funk:proc;
  local i,ihi,ilo,inhi,j,ndim,mpts,tiny,tinyp;
  local rtol,sum,swap,ysave,ytry,psum,tmp;

  tiny=1e-3;
  tinyp=1e-3;
  ndim=rows(p);
  mpts=ndim+1;
  psum=sumc(p');
  do while (1);
    if (maxc(maxc(abs(p-psum/mpts)))<tinyp); break; endif;
    ilo=minindc(y);
    ihi=maxindc(y);
    tmp=y; tmp[ihi]=tmp[ilo];
    inhi=maxindc(tmp);
    rtol=abs(y[ihi]-y[ilo]);
    if (rtol < ftol);
      tmp=y[1];     y[1]=y[ilo];     y[ilo]=tmp;
      tmp=p[.,1]; p[.,1]=p[.,ilo]; p[.,ilo]=tmp;
      break;
    endif;
    {ytry,p,y,psum}=amotry(p,y,psum,ihi,-1.0,&funk,data1,data2);
    if (ytry <= y[ilo]);
      {ytry,p,y,psum}=amotry(p,y,psum,ihi,2.0,&funk,data1,data2);
    elseif (ytry >= y[inhi]);
      ysave=y[ihi];
      {ytry,p,y,psum}=amotry(p,y,psum,ihi,0.5,&funk,data1,data2);
      if (ytry >= ysave);
        i=1;
        do while (i<=mpts);
          if (i /= ilo);
            psum=0.5*(p[.,i]+p[.,ilo]);
            p[.,i]=psum;
          endif;
          i=i+1;
        endo;
        psum=sumc(p');
      endif;
    endif;
  endo;
  retp(psum/mpts);
endp;

/* main function to call for Nelder-Mead algorithm */
/* p0: starting point
   ptol: initial guess of the size of the region containing the minimum
   ftol: target precision for stopping rule
   funk: pointer to function to optimize
   data1: \ additional input parameters to send to funk but not optimize over
   data2: / (data1 will contain the dataset while data2 will contain parameter vectors ) */
proc(1)=amoeba(p0,ptol,ftol,funk,data1,data2);
  local funk:proc;
  local ndim,p,y,i,d;
  ndim=rows(p0);
  p=zeros(ndim,ndim+1);
  p[.,1]=p0;
  y=zeros(ndim+1,1);
  i=0;
  do while (i<=ndim);
    d=zeros(ndim,1);
    if (i>0); d[i]=1.0; endif;
    p[.,i+1]=p0+d*ptol;
    y[i+1]=funk(p[.,i+1],data1,data2);
    i=i+1;
  endo;
  retp(amoeba_loop(p,y,ftol,&funk,data1,data2));
endp;

/* calculates the objective function */
proc(1)=calc_lnL_pre_elvis(ob,theta,gam,rep,tol,myseed);
  local gs;
  gs=avg_mom(&mom,ob,theta,gam,rep,myseed);
  retp(calc_lnL_el(gs,tol));
endp;

/* utility routine to unpack parameters and dataset from data1 and data2 matrices and call objective function evaluation*/
proc(1)=unpack_funk(p,data1,data2);
  local dimtheta,totdim,alltheta,nuis,gam,rep,tol,myseed,nbnuis,v;
  dimtheta=data2[1];
  nbnuis=getnbnuis();
  totdim=rows(data2);

  if (nbnuis==0);
    nuis={};
  else;
    nuis=p[1:nbnuis];
  endif;
  gam=p[nbnuis+1:rows(p)];
  alltheta=data2[2:2+dimtheta-1]|nuis;
  rep=data2[totdim-3:totdim-2];
  tol=data2[totdim-1];
  myseed=data2[totdim];
  v=-calc_lnL_pre_elvis(data1,alltheta,gam,rep,tol,myseed);
  /*if (oxmpi_ismaster()); -v~nuis'~gam'; endif;*/ /*uncomment if you want to trace optimization */
  retp(v);
endp;

/* utility routine to pack parameters into data2 matrix */
proc(1)=pack_funk(theta,rep,tol,myseed);
  retp(rows(theta)|theta|rep|tol|myseed);
endp;

/* main calling point to calculate the objective function */
/* ob: the whole sample of observables (each of the observables are in a row, individuals are stacked in cols)
   theta: parameter vector, excluding nuisance parameters
   nuisgamhint: initial guess of the nuisance parameters and gamma, stacked as a single vector
                if only the nuisance parameters are given, a guess for gamma will be calculated internally
   ptol: initial guess of the size of the region containing the minimum
   ftol: target precision for stopping rule of nuis+gamma optimization
   rep: array controling replications:
    rep[1]: number of equilibration steps
    rep[2]: number of averaging steps
   tol: target precision for stopping rule for EL calculation
   myseed: random number seed
returns:
   1) maximum of objective function and
   2) values of nuisance parameter and gamma
*/
proc(2)=calc_lnL_elvis(ob,theta,nuisgamhint,ptol,ftol,rep,tol,myseed);
  local ob1;
  local gam,gamref;
  local nuis,nuisgam,nbnuis;
  local i;

  nbnuis=getnbnuis();
  if (rows(nuisgamhint)>nbnuis);
    nuisgam=nuisgamhint;
  else;
    ob1=ob[1,.]';
    gam=0*mom(guess_un(ob1,theta|zeros(nbnuis,1),0),ob1,theta|zeros(nbnuis,1));
    nuisgam=nuisgamhint|gam;
  endif;

  nuisgam=amoeba(nuisgam,ptol,ftol,&unpack_funk,ob,pack_funk(theta,rep,tol,myseed));
  if (nbnuis==0);
    nuis={};
  else;
    nuis=nuisgam[1:nbnuis];
  endif;
  gam=nuisgam[nbnuis+1:rows(nuisgam)];
  retp(calc_lnL_pre_elvis(ob,theta|nuis,gam,rep,tol,myseed),nuisgam);
endp;
